// Filename   : ubx_router.cpp
// Description: forwards complete uBlox F9P UBX-NAV-PVT packets via USART1 on JST10-7 pin at 115200 baud
// Author     : Dr. Volodymyr Tymchyshyn, 2024 (c)

#include "AP_Periph.h"

#ifdef UBX_NAV_PVT_ROUTER

#include <stdio.h>
#include <hal.h>

extern const AP_HAL::HAL &hal;



// u-blox F9 HPG L1L5 1.40 - Interface description
// Structure of UBX-NAV-PVT message, Ch. "3.15.13.1 Navigation position velocity time solution"
#define UBX_NAV_PVT_PAYLOAD_LENGTH (92u)
#define UBX_NAV_PVT_PACKET_LENGTH  (UBX_NAV_PVT_PAYLOAD_LENGTH+8u)
//
#define SYNC_CHAR1      0xB5u
#define SYNC_CHAR2      0x62u
#define UBX_NAV_CLASS   0x01u
#define UBX_NAV_PVT_ID  0x07u
//
//
// Valid UBX_NAV_PVT packet must always start with these hardcoded 6 bytes:
static const uint8_t start_UBX_NAV_PVT[6] =
{
 SYNC_CHAR1, SYNC_CHAR2,      // Header
 UBX_NAV_CLASS,               // class
 UBX_NAV_PVT_ID,              // id
 92, 0                        // length, LE
};


class UBX_Router
{
public:
    UBX_Router(uint8_t _serial_num) : serial_num(_serial_num) {}
    void init(void);

private:
    uint8_t serial_num;
    void ubx_rawdata_cb(const uint8_t *data, uint32_t length);
};
//
static UBX_Router ubx_router_uart1(0); // serial0Driver is USART1 on 10-pin connector (with 2 CANx) pin7 is output to FMU sniffing


void UBX_Router::init(void)
{
    auto *uart = hal.serial(serial_num);
    if ( (uart == nullptr) || (!uart->is_initialized()) )
    {
        return;
    }
    // Check the used baudrate, and set to 115200 with 1KB output buffer - if not preset as expected (also in future Here4 code base)...
    const uint32_t baud_rate = uart->get_baud_rate();
    if (baud_rate != 115200)
    {
        uart->begin(AP_SERIALMANAGER_CONSOLE_BAUD, 128, 1024); // Rx, Tx
        if (uart->get_baud_rate() != 115200)
        {
            printf("UBX_Router::init: failed to set 115200 for serial %u.\n", serial_num);
            return;
        }
    }

    // set callback for GPS RAW data
    AP::gps().set_gps_raw_cb(FUNCTOR_BIND_MEMBER(&UBX_Router::ubx_rawdata_cb, void, const uint8_t*, uint32_t));

    printf("\nUBX_Router started at fixed 115200 baud on hal.serial(%u)\n", serial_num);
    can_printf("UBX_NAV_PVT router started");
}


void UBX_Router::ubx_rawdata_cb(const uint8_t *data, uint32_t length)
{
    // currently optimized for routing only single packet type: UBX_NAV_PVT
    if ( (length == UBX_NAV_PVT_PACKET_LENGTH) && (memcmp(data, start_UBX_NAV_PVT, 6) == 0) )
    {
        auto *uart = hal.serial(serial_num);
        if (uart  && uart->is_initialized())
        {
           // serial0Driver USART1 is on JST10 connector (originally "console"). On serial_manager.init() exit: was set to 115200 baud, Rx 128 / Tx 512 with DMA
           uart->write(data, UBX_NAV_PVT_PACKET_LENGTH);
        }

        // Log once per 10s on JST6.FMU_TX (SD6, ChibiOS console @ 57600 baud, no DMA) -
        // to monitor patched HERE4 health and track the number / rate of UBX_NAV_PVT messages
        static size_t num_packets_routed = 0;
        static size_t last_reported_ms   = 0;
        num_packets_routed++;
        const size_t now = AP_HAL::millis();
        if ( (now - last_reported_ms) > 9950) // called each 10s regardless of message Rate (5Hz or 10Hz, etc.)
        {
            can_printf("UBX_NAV_PVT router: %u", unsigned(num_packets_routed));
            printf("UBX_NAV_PVT router: %u packets.\n", num_packets_routed);
            last_reported_ms = now;
        }
    }
}

extern void UBX_Router_install(void);
void UBX_Router_install(void)
{
    ubx_router_uart1.init();
}

#endif // UBX_NAV_PVT_ROUTER
